(*  
    Author:      René Thiemann 
                 Akihisa Yamada
    License:     BSD
*)
section \<open>Explicit Roots\<close>

text \<open>We provide algorithms which use the explicit formulas to 
  compute the roots of polynomials of degree up to 2. We also define the formula for 
  polynomials of degree 3, but did not (yet) prove anything about it, as it will not be 
  executable.

  There are also safe partial variants of the root finding algorithms which take into
  account that not all operations are executable. For instance the roots of $x^2 - \sqrt{2}$
  which are $2^{1/4}$ and $-2^{1/4}$ are currently not representable in the implementation
  of the real numbers.\<close>  

theory Complex_Roots
imports   
  Missing_Polynomial
  "~~/src/HOL/Library/Fundamental_Theorem_Algebra"
  Complex_Impl
begin

section {* Roots of polynomials *}

lemma degree0_coeffs: "degree p = 0 \<Longrightarrow>
  \<exists> a. p = [: a :]"
  by (metis degree_pCons_eq_if old.nat.distinct(2) pCons_cases)

lemma degree1_coeffs: "degree p = 1 \<Longrightarrow>
  \<exists> a b. p = [: b, a :] \<and> a \<noteq> 0" 
  by (metis One_nat_def degree_pCons_eq_if nat.inject old.nat.distinct(2) pCons_0_0 pCons_cases)

lemma degree2_coeffs: "degree p = 2 \<Longrightarrow>
  \<exists> a b c. p = [: c, b, a :] \<and> a \<noteq> 0"
  by (metis Suc_1 Suc_neq_Zero degree1_coeffs degree_pCons_eq_if nat.inject pCons_cases)

lemma roots0: assumes p: "p \<noteq> 0" and p0: "degree p = 0" 
  shows "{x. poly p x = 0} = {}"
  using degree0_coeffs[OF p0] p by auto

definition roots1 :: "'a :: field poly \<Rightarrow> 'a" where
  "roots1 p = (- coeff p 0 / coeff p 1)"

lemma roots1: fixes p :: "'a :: field poly"
  assumes p1: "degree p = 1" 
  shows "{x. poly p x = 0} = {roots1 p}"
  using degree1_coeffs[OF p1] unfolding roots1_def 
  by (auto simp: add_eq_0_iff nonzero_neg_divide_eq_eq2)

definition croots2 :: "complex poly \<Rightarrow> complex list" where
  "croots2 p = (let a = coeff p 2; b = coeff p 1; c = coeff p 0; b2a = b / (2 * a);
    bac = b2a^2 - c/a;
    e = csqrt bac 
    in
     remdups [- b2a + e, - b2a - e])"

definition complex_rat :: "complex \<Rightarrow> bool" where
  "complex_rat x = (Re x \<in> \<rat> \<and> Im x \<in> \<rat>)"

text \<open>Ensure that @{const csqrt} will not require computation of
  square roots of irrational numbers. (This is not supported by the
  current implementation.)\<close>

definition complex_sqrt_able :: "complex \<Rightarrow> bool" where
  "complex_sqrt_able x = (complex_rat x \<and> complex_rat (cmod x))"

definition croots2_no_sqrt :: "complex poly \<Rightarrow> complex list" where
  "croots2_no_sqrt p = (let a = coeff p 2; b = coeff p 1; c = coeff p 0; b2a = b / (2 * a);
    bac = b2a^2 - c/a
  in if complex_sqrt_able bac then
    let e = csqrt bac 
    in
     remdups [- b2a + e, - b2a - e] else [])"

lemma croots2: assumes "degree p = 2"
  shows "{x. poly p x = 0} = set (croots2 p)"
proof -
  from degree2_coeffs[OF assms] obtain a b c 
  where p: "p = [:c, b, a:]" and a: "a \<noteq> 0" by auto
  have 2: "2 = Suc (Suc 0)" by simp
  have coeff: "coeff p 2 = a" "coeff p 1 = b" "coeff p 0 = c" unfolding p by (auto simp: 2)
  have poly: "\<And> x. poly p x = x * x * a + x * b + c" unfolding p by (auto simp: field_simps)
  let ?b2a = "b / (2 * a)"
  def b2a \<equiv> ?b2a
  let ?bac = "b2a^2 - c/a"
  def bac \<equiv> ?bac
  have roots: "set (croots2 p) = {- b2a + csqrt bac, - b2a - csqrt bac}"
    unfolding croots2_def Let_def coeff b2a_def[symmetric] bac_def[symmetric]
    by (auto split: if_splits)
  {
    fix x :: complex
    have "(poly p x = 0) = (x * x * a + x * b + c = 0)" unfolding poly ..
    also have "\<dots> = ((x * x + 2 * x * b2a) + c/a = 0)" using a by (auto simp: b2a_def field_simps)
    also have "x * x + 2 * x * b2a = (x * x + 2 * x * b2a + b2a^2) - b2a^2" by simp
    also have "\<dots> = (x + b2a) ^ 2 - b2a ^ 2" 
      by (simp add: field_simps power2_eq_square)
    also have "(\<dots> + c / a = 0) = ((x + b2a) ^ 2 = bac)" unfolding bac_def by algebra
    also have "\<dots> = (x + b2a = csqrt bac \<or> -(x + b2a) = csqrt bac)"
      by (metis power2_csqrt power2_eq_iff)
    also have "(x + b2a = csqrt bac) = (x = - b2a + csqrt bac)"
      by (metis diff_eq_eq uminus_add_conv_diff)
    also have "(-(x + b2a) = csqrt bac) = (x = -b2a - csqrt bac)"
      by (metis add.commute diff_add_cancel diff_minus_eq_add)
    finally have "(poly p x = 0) = (x = -b2a + csqrt bac \<or> x = -b2a - csqrt bac)" .
  } note main = this
  show ?thesis unfolding roots main by auto
qed

definition rroots2 :: "real poly \<Rightarrow> real list" where
  "rroots2 p = (let a = coeff p 2; b = coeff p 1; c = coeff p 0; b2a = b / (2 * a);
    bac = b2a^2 - c/a
  in if bac = 0 then [- b2a] else if bac < 0 then []
    else let e = sqrt bac
    in
     [- b2a + e, - b2a - e])"

lemma croots2_no_sqrt: assumes d2: "degree p = 2"
  shows 
  "set (croots2_no_sqrt p) \<subseteq> {x. poly p x = 0}"
  "a = coeff p 2 \<Longrightarrow> b = coeff p 1 \<Longrightarrow> c = coeff p 0 \<Longrightarrow> complex_sqrt_able ((b / (2 * a))^2 - c / a) \<Longrightarrow>
    set (croots2_no_sqrt p) = {x. poly p x = 0}"
  unfolding croots2[OF d2]
  unfolding croots2_def Let_def croots2_no_sqrt_def 
  by auto

definition rroots2_no_sqrt :: "real poly \<Rightarrow> real list" where
  "rroots2_no_sqrt p = (let a = coeff p 2; b = coeff p 1; c = coeff p 0; b2a = b / (2 * a);
    bac = b2a^2 - c/a
  in if bac = 0 then [- b2a] else if bac < 0 \<or> bac \<notin> \<rat> then []
    else let e = sqrt bac 
    in
     [- b2a + e, - b2a - e])"

lemma rroots2: assumes "degree p = 2"
  shows "{x. poly p x = 0} = set (rroots2 p)"
proof -
  from degree2_coeffs[OF assms] obtain a b c 
  where p: "p = [:c, b, a:]" and a: "a \<noteq> 0" by auto
  have 2: "2 = Suc (Suc 0)" by simp
  have coeff: "coeff p 2 = a" "coeff p 1 = b" "coeff p 0 = c" unfolding p by (auto simp: 2)
  have poly: "\<And> x. poly p x = x * x * a + x * b + c" unfolding p by (auto simp: field_simps)
  let ?b2a = "b / (2 * a)"
  def b2a \<equiv> ?b2a
  let ?bac = "b2a^2 - c/a"
  def bac \<equiv> ?bac
  have roots: "set (rroots2 p) = (if bac < 0 then {} else {- b2a + sqrt bac, - b2a - sqrt bac})"
    unfolding rroots2_def Let_def coeff b2a_def[symmetric] bac_def[symmetric]
    by (auto split: if_splits)
  {
    fix x :: real
    have "(poly p x = 0) = (x * x * a + x * b + c = 0)" unfolding poly ..
    also have "\<dots> = ((x * x + 2 * x * b2a) + c/a = 0)" using a by (auto simp: b2a_def field_simps)
    also have "x * x + 2 * x * b2a = (x * x + 2 * x * b2a + b2a^2) - b2a^2" by simp
    also have "\<dots> = (x + b2a) ^ 2 - b2a ^ 2" 
      by (simp add: field_simps power2_eq_square)
    also have "(\<dots> + c / a = 0) = ((x + b2a) ^ 2 = bac)" unfolding bac_def by linarith
    also have "\<dots> = (bac \<ge> 0 \<and> ((x + b2a) = sqrt bac \<or> -(x + b2a) = sqrt bac))"
      by (metis power2_eq_iff real_sqrt_pow2 zero_le_power2)
    also have "\<dots> = (bac \<ge> 0 \<and> (x = -b2a + sqrt bac \<or> x = -b2a - sqrt bac))" by linarith      
    finally have "(poly p x = 0) = (bac \<ge> 0 \<and> (x = -b2a + sqrt bac \<or> x = -b2a - sqrt bac))" .
  } note main = this
  show ?thesis unfolding roots main by auto
qed

lemma rroots2_no_sqrt: assumes d2: "degree p = 2"
  shows 
  "set (rroots2_no_sqrt p) \<subseteq> {x. poly p x = 0}"
  "a = coeff p 2 \<Longrightarrow> b = coeff p 1 \<Longrightarrow> c = coeff p 0 \<Longrightarrow> (b / (2 * a))^2 - c / a \<in> \<rat> \<Longrightarrow>
  set (rroots2_no_sqrt p) = {x. poly p x = 0}"
  unfolding rroots2[OF d2]
  unfolding rroots2_def Let_def rroots2_no_sqrt_def 
  by auto


text \<open>Determinining roots of complex polynomials of degree up to 2.
  We use @{const croots2_no_sqrt} so that the algorithm will yield no
  runtime errors.\<close>

definition croots :: "complex poly \<Rightarrow> complex list" where
  "croots p = (if p = 0 \<or> degree p > 2 then [] 
    else (if degree p = 0 then [] else if degree p = 1 then [roots1 p]
    else croots2 p))"

definition croots_no_sqrt :: "complex poly \<Rightarrow> complex list" where
  "croots_no_sqrt p = (if p = 0 \<or> degree p > 2 then [] 
    else (if degree p = 0 then [] else if degree p = 1 then [roots1 p]
    else croots2_no_sqrt p))"

lemma croots: assumes "p \<noteq> 0" "degree p \<le> 2"
  shows "set (croots p) = {x. poly p x = 0}"
  using assms unfolding croots_def
  using roots0[of p] roots1[of p] croots2[of p]
  by (auto split: if_splits)

lemma croots_no_sqrt: assumes "p \<noteq> 0" "degree p \<le> 2"
  shows "set (croots_no_sqrt p) \<subseteq> {x. poly p x = 0}"
  using assms unfolding croots_no_sqrt_def
  using roots0[of p] roots1[of p] croots2_no_sqrt[of p]
  by (auto split: if_splits)

text \<open>Determinining roots of real polynomials of degree up to 2.\<close>

definition rroots :: "real poly \<Rightarrow> real list" where
  "rroots p = (if p = 0 \<or> degree p > 2 then [] 
    else (if degree p = 0 then [] else if degree p = 1 then [roots1 p]
    else rroots2 p))"

definition rroots_no_sqrt :: "real poly \<Rightarrow> real list" where
  "rroots_no_sqrt p = (if p = 0 \<or> degree p > 2 then [] 
    else (if degree p = 0 then [] else if degree p = 1 then [roots1 p]
    else rroots2_no_sqrt p))"

lemma rroots: assumes "p \<noteq> 0" "degree p \<le> 2"
  shows "set (rroots p) = {x. poly p x = 0}"
  using assms unfolding rroots_def
  using roots0[of p] roots1[of p] rroots2[of p]
  by (auto split: if_splits)

lemma rroots_no_sqrt: assumes "p \<noteq> 0" "degree p \<le> 2"
  shows "set (rroots_no_sqrt p) \<subseteq> {x. poly p x = 0}"
  using assms unfolding rroots_no_sqrt_def
  using roots0[of p] roots1[of p] rroots2_no_sqrt[of p]
  by (auto split: if_splits)


text \<open>Although there is a closed form for cubic roots, 
  which is specified below, we did not integrate it into the
 @{const croots} and @{const rroots} algorithms.
 The reason is that currently cubic roots on complex and real numbers
 are not executable. Moreover, for complex numbers, the cubic root is not
 even defined. Therefore, we also did not proof soundness of the croots3 algorithm.\<close>

definition croot :: "nat \<Rightarrow> complex \<Rightarrow> complex" where 
  "croot n c = (SOME d. d^n = c)"

definition croots3 :: "complex poly \<Rightarrow> complex \<times> complex \<times> complex \<times> complex" where
  "croots3 p = (let a = coeff p 3; b = coeff p 2; c = coeff p 1; d = coeff p 0; 
    \<Delta>\<^sub>0 = b^2 - 3 * a * c;
    \<Delta>\<^sub>1 = 2 * b^3 - 9 * a * b * c + 27 * a^2 * d;
    C = croot 3 ((\<Delta>\<^sub>1 + csqrt ( \<Delta>\<^sub>1^2 - 4 * \<Delta>\<^sub>0^3)) / 2);
    u\<^sub>1 = 1;
    u\<^sub>2 = (-1 + \<i> * csqrt 3) / 2;
    u\<^sub>3 = (-1 - \<i> * csqrt 3) / 2;
    x\<^sub>k = (\<lambda> u. (-1 / (3 * a)) * (b + u * C + \<Delta>\<^sub>0 / (u * C)))
    in
     (x\<^sub>k u\<^sub>1, x\<^sub>k u\<^sub>2, x\<^sub>k u\<^sub>3, a))"

lemma croot_pow: assumes n: "n > 0" shows "(croot n c)^n = c"
proof -
  from n obtain k where n: "n = Suc k" by (cases n, auto)
  show ?thesis 
    unfolding croot_def
  proof (rule someI_ex)
    def p \<equiv> "[: - c :] + ([: 0, 1 :] ^ n)"
    have "\<not> constant (poly p)" unfolding constant_def
    proof (rule, elim allE)
      assume "poly p 0 = poly p 1"
      thus False unfolding p_def n by simp
    qed
    from fundamental_theorem_of_algebra[OF this] obtain a where a: "poly p a = 0" by auto
    from this[unfolded p_def] have "a ^ n = c" by auto
    thus "\<exists> x. x ^ n = c" by auto
  qed
qed

end