(*  
    Author:      René Thiemann 
                 Akihisa Yamada
    License:     BSD
*)
section \<open>Checking Polynomial Growth of Matrices\<close>

text \<open>We combine the developed algorithms into a single check-function which
  tries to ensure polynomial growth of a matrix w.r.t. to a given degree.\<close>
 
theory Jordan_Normal_Form_Complexity_Approximation
imports
  Jordan_Normal_Form_Uniqueness
  Jordan_Normal_Form_Existence
  Polynomial_Conversion
  Matrix_Complexity
  Matrix_Show
  Complex_Show
  Poly_Show
  Root_Finder
  "$AFP/Sturm_Sequences/Sturm_Theorem"
  "$AFP/Certification_Monads/Check_Monad"
begin

definition csample_values :: "complex list" where "csample_values = [1,-1,1/2,0,-1/2,-\<i>,\<i>]"

definition csample_pvalues :: "complex poly list" where "csample_pvalues = [[:1,1,1:]]"

definition determine_roots :: "complex \<times> (complex poly \<times> nat)list \<Rightarrow> complex list \<times> complex poly list" where
  "determine_roots aps \<equiv> case aps of (a,ps) \<Rightarrow> 
    (case partition (\<lambda> p. degree (fst p) = 1) ps of (deg1,other) \<Rightarrow>
    (map (roots1 o fst) deg1, map fst other))"

lemma determine_roots: assumes sff: "square_free_factorization p aps"
  and p0: "p \<noteq> 0"
  and res: "determine_roots aps = (rts, qs)"
  shows "{ x. poly p x = 0 } = set rts \<union> { x. \<exists> q \<in> set qs. poly q x = 0}"
proof -
  obtain a ps where aps: "aps = (a,ps)" by force
  let ?f = "\<lambda> p. degree (fst p) = 1"
  obtain deg1 other where part: "partition ?f ps = (deg1, other)" by force
  from partition_P[OF part] have deg1: "\<And> q. q \<in>set deg1 \<Longrightarrow> degree (fst q) = 1" by auto
  from res[unfolded determine_roots_def aps split part] have rts: "rts = (map (roots1 o fst) deg1)"
    and qs: "qs = map fst other" by auto
  from square_free_factorizationD(1,4)[OF sff[unfolded aps]]
  have p: "p = Polynomial.smult a (\<Prod>(a, i)\<in>set ps. a ^ Suc i)" by auto
  from square_free_factorization_root[OF sff[unfolded aps] p0]
  have "{x. poly p x = 0} = {x. \<exists>a i. (a, i) \<in> set deg1 \<union> set other \<and> poly a x = 0}" by (auto simp: partition_set[OF part])
  also have "\<dots> = {x. \<exists>a i. (a, i) \<in> set deg1 \<and> poly a x = 0} \<union> {x . \<exists> a i. (a,i) \<in> set other \<and> poly a x = 0}" (is "_ = ?R1 \<union> ?Ro") by blast
  also have "?Ro = { x. \<exists> q \<in> set qs. poly q x = 0}" unfolding qs by force
  also have "?R1 = (roots1 \<circ> fst) ` set deg1" (is "_ = ?RR1")
  proof -
    have "?RR1 \<subseteq> ?R1" using roots1 deg1 by force
    moreover
    {
      fix x
      assume "x \<in> ?R1"
      then obtain a i where ai: "(a,i) \<in> set deg1" and z: "poly a x = 0" by auto
      with roots1[OF deg1[OF ai]] have "x \<in> ?RR1" by force
    }
    ultimately show ?thesis by blast
  qed
  also have "\<dots> = set rts" unfolding rts by auto
  finally show ?thesis by simp
qed

definition analyse_evs :: "nat \<Rightarrow> complex mat \<Rightarrow> complex list \<Rightarrow> shows check" where
  "analyse_evs d A evs \<equiv> check_allm (\<lambda> e. do {
       check (is_norm_le_1 e) (shows ''detected eigenvalue with norm > 1: '' o shows e);
       if is_norm_1 e then 
         check_allm (\<lambda> (n,_). check (n \<le> d) (shows ''detected Jordan-Block of size '' o shows n
           o shows '' for eigenvalue '' o shows e)) (compute_set_of_jordan_blocks A e)
       else succeed
     }) evs"


definition count_roots_between_open :: "real poly \<Rightarrow> real \<Rightarrow> real \<Rightarrow> nat" where
  "count_roots_between_open p a b = count_roots_between p a b - 
    (if a < b \<and> poly p b = 0 then 1 else 0)"
  
lemma count_roots_between_open: "count_roots_between_open p a b = 
  card ({ x. a < x \<and> x < b \<and> poly p x = 0})"
proof -
  have id: "{ x. a < x \<and> x < b \<and> poly p x = 0} = { x. a < x \<and> x \<le> b \<and> poly p x = 0} - {b}" 
    (is "_ = ?R - _") by auto  
  show ?thesis  
    by (cases "finite ?R", 
      auto simp: id count_roots_between_correct count_roots_between_open_def)
qed
 
definition ensure_small_real_roots :: "complex poly \<Rightarrow> bool" where
  "ensure_small_real_roots p = (p \<noteq> 0 \<and> (\<forall> x \<in> set (coeffs p). x \<in> \<real>)
  \<and> count_roots_between_open (map_poly Re p) (-1) 1 = degree p)"

lemma ensure_small_real_roots: assumes ens: "ensure_small_real_roots p"
  shows "{x. poly p x = 0} \<subseteq> complex_of_real ` { x. -1 < x \<and> x < 1}" 
proof -
  let ?p = "map_poly Re p"
  from ens[unfolded ensure_small_real_roots_def Let_def] have 
    p0: "p \<noteq> 0" 
  and real: "set (coeffs p) \<subseteq> \<real>" 
  and count: "count_roots_between_open ?p (-1) 1 = degree p" 
    by blast+
  from poly_roots_degree[OF p0, folded count, unfolded count_roots_between_open]
  have "card {x. poly p x = 0} \<le> card {x. - 1 < x \<and> x < 1 \<and> poly ?p x = 0}" .
  also have "\<dots> = card (complex_of_real ` {x. - 1 < x \<and> x < 1 \<and> poly ?p x = 0})"
    by (rule card_image[symmetric], auto simp: inj_on_def)
  finally have le: "card {x. poly p x = 0} \<le> card (complex_of_real ` {x. - 1 < x \<and> x < 1 \<and> poly ?p x = 0})" 
    (is "card ?A \<le> card ?B") by auto
  have "?B = ?A"
  proof (rule card_seteq[OF _ _ le])
    show "finite ?A" using p0 by (rule poly_roots_finite)
    {
      fix c
      assume "c \<in> ?B"
      then obtain x where c: "c = of_real x" and px: "poly ?p x = 0" by auto
      from px have "poly p c = 0" using real unfolding c by (metis map_poly_Re_poly of_real_0)
    }
    thus "?B \<subseteq> ?A" by auto
  qed
  thus ?thesis by auto
qed

definition analyse_polys :: "complex poly list \<Rightarrow> shows check" where
  "analyse_polys polys \<equiv> check_allm (\<lambda> p. do {
       check (ensure_small_real_roots p) (shows ''could not ensure that eigenvalues of '' o shows p o shows '' are small'')
     }) polys"

definition approximate_jordan_blocks_complex_mat :: "complex list \<Rightarrow> complex poly list \<Rightarrow> nat \<Rightarrow> complex mat \<Rightarrow> shows check" where
  "approximate_jordan_blocks_complex_mat csamples cpsamples d A \<equiv> let
     n = dim\<^sub>r A;
     c_samples = remdups (csample_values @ csamples @ (remdups ([A $$ (i,j) . i <- [0 ..< n], j <- [0 ..< n]])));
     cp_samples = remdups (cpsamples @ csample_pvalues);
     cp = char_poly A;
     sff = complex_factorization c_samples cp_samples cp;
     (evs, polys) = determine_roots sff
     in do {
       analyse_evs d A evs;
       analyse_polys polys
     } <+? (\<lambda> s. shows ''problem in estimating growth rate of matrix\<newline>'' o shows A o 
       shows ''\<newline>with characteristic polynomial '' o shows cp o shows_nl o s)"

lemma approximate_jordan_blocks_complex_mat: assumes A: "A \<in> carrier\<^sub>m n n"
  and res: "isOK(approximate_jordan_blocks_complex_mat csamples cpsamples d A)"
shows "\<exists>c1 c2. \<forall>k. norm_bound (A ^\<^sub>m k) (c1 + c2 * of_nat k ^ (d - 1))" 
proof -
  from char_poly_factorized[OF A] obtain n_as' where "char_poly A = (\<Prod>a\<leftarrow>n_as'. [:- a, 1:])" by auto
  from jordan_nf_exists[OF A this] obtain n_as where jnf: "jordan_nf A n_as" by auto
  from jordan_nf_char_poly[OF this] have cp: "char_poly A = (\<Prod>(n, a)\<leftarrow>n_as. [:- a, 1:] ^ n)" .
  from res
  obtain evs polys c_samples cp_samples where dr: "determine_roots (complex_factorization c_samples cp_samples (char_poly A)) = (evs, polys)" 
    and ana_evs: "isOK(analyse_evs d A evs)" 
    and ana_polys: "isOK(analyse_polys polys)"
    unfolding approximate_jordan_blocks_complex_mat_def Let_def by (auto split: prod.splits)
  let ?cp = "char_poly A"
  let ?rpl = "{x. \<exists>q\<in>set polys. poly q x = 0}"
  from degree_monic_char_poly[OF A] have "?cp \<noteq> 0" by auto
  from determine_roots[OF complex_factorization this dr]
  have rts: "{x. poly ?cp x = 0} = set evs \<union> ?rpl" by auto
  show ?thesis
  proof (rule jordan_nf_matrix_poly_bound[OF A _ _ jnf])
    fix n a
    assume na: "(n,a) \<in> set n_as"
    then obtain as bs where nas: "n_as = as @ (n,a) # bs" by (meson split_list)
    obtain q where cp: "char_poly A = ([:-a,1:]^n) * q" unfolding cp nas by auto
    { 
      assume n: "0 < n"
      with cp na have rt: "poly ?cp a = 0" by auto
      with rts have "a \<in> set evs \<or> a \<in> ?rpl" by auto
      hence "norm a \<le> 1 \<and> (norm a = 1 \<longrightarrow> n \<le> d)"
      proof
        assume a: "a \<in> set evs" 
        let ?ns = "[n . ev <- evs, norm ev = 1, n <- map fst (compute_set_of_jordan_blocks A ev)]"
        from ana_evs[unfolded analyse_evs_def] a have 
          cb: "\<And> n b . norm a = 1 \<Longrightarrow> (n,b) \<in> set (compute_set_of_jordan_blocks A a) \<Longrightarrow> n \<le> d" and 
          norm_a: "norm a \<le> 1" 
          by auto
        {
          assume n1: "norm a = 1"
          have "n \<le> d"
            by (rule cb[OF n1], unfold compute_set_of_jordan_blocks[OF jnf], insert n na a, auto)
        }
        with norm_a show ?thesis by auto
      next
        assume "a \<in> ?rpl"
        then obtain q where q: "q \<in> set polys" and qa: "poly q a = 0" by auto
        from ana_polys[unfolded analyse_polys_def] q have "ensure_small_real_roots q" by auto
        from ensure_small_real_roots[OF this] qa show ?thesis by auto
      qed
    } note * = this
    show "0 < n \<Longrightarrow> cmod a \<le> 1" using * by auto
    show "cmod a = 1 \<Longrightarrow> n \<le> d" using * by (cases "n = 0", auto)
  qed
qed

definition count_ones_check :: "real mat \<Rightarrow> nat \<Rightarrow> bool" where
  "count_ones_check A d \<equiv> upper_triangular A \<and> (let diag = mat_diag A
     in (\<forall> a \<in> set diag. abs a \<le> 1 \<and> (abs a = 1 \<longrightarrow> length (filter (op = a) diag) \<le> d)))"

text \<open>We first try a cheap test whether we have the simple case: an upper-triangular matrix. 
  If the counting-ones criterion succeeds, then we are done, and only otherwise we start the
  full machinery with root-finding algorithms, Jordan-block computation, etc.\<close>

definition approximate_jordan_blocks_real_mat :: "complex list \<Rightarrow> complex poly list \<Rightarrow> nat \<Rightarrow> real mat \<Rightarrow> shows check" where
  "approximate_jordan_blocks_real_mat csamples cpsamples d A \<equiv>
    if count_ones_check A d then succeed else 
     approximate_jordan_blocks_complex_mat csamples cpsamples d (map\<^sub>m of_real A)"

lemma approximate_jordan_blocks_real_mat: assumes A: "A \<in> carrier\<^sub>m n n"
  and res: "isOK(approximate_jordan_blocks_real_mat csamples cpsamples d A)"
shows "\<exists>c1 c2. \<forall>k. norm_bound (A ^\<^sub>m k) (c1 + c2 * of_nat k ^ (d - 1))" 
proof (cases "count_ones_check A d")
  case False
  let ?B = "map\<^sub>m complex_of_real A"
  from res False have res: "isOK(approximate_jordan_blocks_complex_mat csamples cpsamples d ?B)"
    unfolding approximate_jordan_blocks_real_mat_def by auto
  from A have B: "?B \<in> carrier\<^sub>m n n" by auto
  from approximate_jordan_blocks_complex_mat[OF B res]
  obtain c1 c2 where bnd: "\<And> k. norm_bound (?B ^\<^sub>m k) (c1 + c2 * of_nat k ^ (d - 1))" by blast
  show ?thesis
  proof (rule exI[of _ c1], rule exI[of _ c2], intro allI)
    fix k
    let ?c = "(c1 + c2 * real_of_nat k ^ (d - 1))"
    def c \<equiv> ?c
    interpret ring_hom complex_of_real
      by (unfold_locales, auto)
    from bnd[of k] have "norm_bound (?B ^\<^sub>m k) c" unfolding c_def .
    hence "norm_bound (A ^\<^sub>m k) c" unfolding norm_bound_def
      unfolding mat_hom_pow[OF A, symmetric] by auto
    thus "norm_bound (A ^\<^sub>m k) ?c" unfolding c_def .
  qed
next
  case True
  show ?thesis
    by (rule counting_ones_complexity[OF A], insert True[unfolded count_ones_check_def Let_def], auto)
qed

definition mat_estimate_complexity_jb :: "nat \<Rightarrow> 'a :: large_real_ordered_semiring_1 mat \<Rightarrow> shows check" where
  "mat_estimate_complexity_jb d A \<equiv> let B = mat\<^sub>\<real> A in do {
     check (dim\<^sub>r A = dim\<^sub>c A) (shows ''expected square matrix but got \<newline>'' o shows B);
     approximate_jordan_blocks_real_mat [] [] d B
   }"

lemma mat_estimate_complexity_jb_norm_bound: assumes *: "isOK(mat_estimate_complexity_jb d A)"
  shows "\<exists> c1 c2. \<forall> k. norm_bound (mat\<^sub>\<real> A ^\<^sub>m k) (c1 + c2 * of_nat k ^ (d - 1))"
proof -
  note * = *[unfolded mat_estimate_complexity_jb_def Let_def]
  from * obtain n where A: "A \<in> carrier\<^sub>m n n" unfolding mat_carrier_def by auto 
  from * have *: "isOK(approximate_jordan_blocks_real_mat [] [] d (mat\<^sub>\<real> A))" by auto
  let ?B = "mat\<^sub>\<real> A"
  from A have "?B \<in> carrier\<^sub>m n n" by auto
  from approximate_jordan_blocks_real_mat[OF this *] show ?thesis .
qed

context
  fixes A B C :: "'a :: large_real_ordered_semiring_1 mat" and n d e :: nat
  assumes *: "isOK(mat_estimate_complexity_jb e A)"
  and A: "A \<in> carrier\<^sub>m n n" and B: "B \<in> carrier\<^sub>m n n" and C: "C \<in> carrier\<^sub>m n n"
  and d: "d = e - 1"
begin

lemma mat_estimate_complexity_jb_norm_bound_prod: 
  "\<exists> c. \<forall> k. k > 0 \<longrightarrow> norm_bound (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C)) (c * of_nat k ^ d)"
proof -
  let ?R = "mat\<^sub>\<real>"
  let ?A = "?R A"
  let ?B = "?R B"
  let ?C = "?R C"
  from mat_estimate_complexity_jb_norm_bound[OF *, folded d]
  obtain c1 c2 where bnd: "\<And> k. norm_bound (?A ^\<^sub>m k) (c1 + c2 * of_nat k ^ d)" by auto
  def c \<equiv> "(max c1 0) + (max c2 0)"
  from norm_bound_max[of ?B] obtain nb where nb: "norm_bound ?B nb" by auto
  from norm_bound_max[of ?C] obtain nc where nc: "norm_bound ?C nc" by auto
  let ?n = "of_nat n :: real"
  def b \<equiv> "nb * nc * ?n * ?n * c"
  {
    fix k
    assume "k > (0 :: nat)"
    let ?k = "of_nat k :: real"
    from bnd[of k] have "norm_bound (?A ^\<^sub>m k) (c1 + c2 * ?k ^ d)" .
    have "c1 \<le> (max c1 0) * 1 ^ d" by auto
    also have "\<dots> \<le> (max c1 0) * ?k ^ d"
      by (rule mult_left_mono, insert `k > 0`, auto)
    also have "\<dots> + c2 * ?k ^ d \<le> c * ?k ^ d"
      unfolding c_def by (auto simp: field_simps intro: mult_right_mono)
    finally have "c1 + c2 * of_nat k ^ d \<le> c * of_nat k ^ d" by auto
    with bnd[of k] have "norm_bound (?A ^\<^sub>m k) (c * of_nat k ^ d)"
      unfolding norm_bound_def by force
    from norm_bound_mult[OF mat_pow_closed _ this nc]
    have "norm_bound (?A ^\<^sub>m k \<otimes>\<^sub>m ?C) (c * ?k ^ d * nc * ?n)" using A C by auto
    from norm_bound_mult[OF _ mat_mult_mat_closed[OF mat_pow_closed] nb this]
    have "norm_bound (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C)) (b * ?k ^ d)" unfolding b_def using A B C
      by (auto simp: ac_simps)
  } note bnd = this
  thus ?thesis by auto
qed

text \<open>This is the main result for real valued matrices.\<close>

lemma mat_estimate_complexity_jb_norm_mat_sum_prod: 
  "\<exists> c. \<forall> k. k > 0 \<longrightarrow> norm (mat_sum (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C))) \<le> (c * of_nat k ^ d)"
proof -
  let ?R = "mat\<^sub>\<real>"
  let ?A = "?R A"
  let ?B = "?R B"
  let ?C = "?R C"
  from mat_estimate_complexity_jb_norm_bound_prod obtain c
  where bnd: "\<And> k. k > 0 \<Longrightarrow> norm_bound (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C)) (c * of_nat k ^ d)" by auto
  let ?n = "of_nat n :: real"
  def b \<equiv> "?n * ?n * c" 
  {
    fix k
    let ?nn = "{0 ..< n}"
    let ?g = "\<lambda> i. c * of_nat k ^ d"
    assume "k > (0 :: nat)"    
    from bnd[OF this] have bnd: "norm_bound (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C)) (c * of_nat k ^ d)" .
    have "norm (mat_sum (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C))) \<le> setsum ?g (?nn \<times> ?nn)"
      unfolding mat_sum_def
      by (rule order_trans[OF setsum_norm_le[of _ _ ?g]],
      insert bnd A B C, auto simp: norm_bound_def)
    also have "\<dots> = b * of_nat k ^ d" unfolding b_def setsum_constant
      by (simp add: ac_simps of_nat_mult)
    finally have "norm (mat_sum (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C))) \<le> b * of_nat k ^ d" by auto
  }
  thus ?thesis by auto
qed

text \<open>And via conversion, it also holds for matrices over the intended semiring.\<close>

lemma mat_estimate_complexity_jb_mat_sum_prod: 
  "\<exists> c. \<forall> k. k > 0 \<longrightarrow> mat_sum (B \<otimes>\<^sub>m (A ^\<^sub>m k \<otimes>\<^sub>m C)) \<le> (c * of_nat k ^ d)"
proof -  
  from mat_estimate_complexity_jb_norm_mat_sum_prod obtain c where 
    bnd: "\<And> k. k > 0 \<Longrightarrow> norm (mat_sum (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C))) \<le> (c * of_nat k ^ d)"
    by auto
  note mat_real_conv = 
    real_embedding.hom_mat_sum 
    real_embedding.mat_hom_pow [of _ n]
    real_embedding.mat_hom_mult[of _ n n]
  {
    fix k
    assume "k > (0 :: nat)"
    let ?sum = "mat_sum (B \<otimes>\<^sub>m (A ^\<^sub>m k \<otimes>\<^sub>m C))"
    have Ak: "A ^\<^sub>m k \<in> carrier\<^sub>m n n" using A by simp
    have AkC: "A ^\<^sub>m k \<otimes>\<^sub>m C \<in> carrier\<^sub>m n n" using Ak C by simp
    let ?ck = "c * of_nat k ^ d"
    let ?cck = "of_int \<lceil>c\<rceil> * of_nat k ^ d :: real"
    from bnd[OF `k > 0`] have "mat_sum (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C)) \<le> ?ck" by auto
    also have "mat_sum (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C)) = 
      real ?sum" 
      using Ak AkC A B C by (simp add: mat_real_conv)
    finally have "real ?sum \<le> ?ck" .
    from real_le[OF this] have le: "?sum \<le> of_int \<lceil>?ck\<rceil>" by auto
    have "?ck \<le> ?cck"
      by (rule mult_right_mono, auto, linarith)
    hence "\<lceil>?ck\<rceil> \<le> \<lceil>?cck\<rceil>" by linarith
    hence "of_int \<lceil>?ck\<rceil> \<le> ((of_int \<lceil>?cck\<rceil>) :: 'a)" 
      unfolding of_int_le_iff .
    with le have le: "?sum \<le> of_int \<lceil>?cck\<rceil>" by linarith
    also have "\<dots> = of_int \<lceil>c\<rceil> * of_nat k ^ d"
      by (metis (no_types, hide_lams) ceiling_of_int of_int_mult of_int_of_nat_eq of_nat_power)
    finally have "?sum \<le> of_int \<lceil>c\<rceil> * of_nat k ^ d" .
  }
  thus ?thesis by auto
qed
end

end